#ifndef _EGG_CHAN_H
#define _EGG_CHAN_H

typedef struct memstruct {
  char nick[NICKLEN];
  char userhost[UHOSTLEN];
  time_t joined;
  unsigned short flags;
  time_t split;
  time_t last;
  time_t delay;
  struct userrec *user;
  struct memstruct *next;
} memberlist;

#define CHANMETA "#&!+"

#define CHANOP      0x0001
#define CHANVOICE   0x0002
#define FAKEOP      0x0004
#define SENTOP      0x0008
#define SENTDEOP    0x0010
#define SENTKICK    0x0020
#define SENTVOICE   0x0040
#define SENTDEVOICE 0x0080
#define WASOP       0x0100
#define STOPWHO     0x0200
#define FULL_DELAY  0x0400
#define STOPCHECK   0x0800

#define chan_hasvoice(x) (x->flags & CHANVOICE)
#define chan_hasop(x) (x->flags & CHANOP)
#define chan_fakeop(x) (x->flags & FAKEOP)
#define chan_sentop(x) (x->flags & SENTOP)
#define chan_sentdeop(x) (x->flags & SENTDEOP)
#define chan_sentkick(x) (x->flags & SENTKICK)
#define chan_sentvoice(x) (x->flags & SENTVOICE)
#define chan_sentdevoice(x) (x->flags & SENTDEVOICE)
#define chan_issplit(x) (x->split > 0)
#define chan_wasop(x) (x->flags & WASOP)
#define chan_stopcheck(x) (x->flags & STOPCHECK)

typedef struct maskstruct {
  char *mask;
  char *who;
  time_t timer;
  struct maskstruct *next;
} masklist;

typedef struct maskrec {
  struct maskrec *next;
  char *mask,
       *desc,
       *user;
  time_t expire,
         added,
         lastactive;
  int flags;
} maskrec;
extern maskrec *global_bans, *global_exempts, *global_invites;

#define MASKREC_STICKY 1
#define MASKREC_PERM   2

struct chan_t {
  memberlist *member;
  masklist *ban;
  masklist *exempt;
  masklist *invite;
  char *topic;
  char *key;
  unsigned short int mode;
  int maxmembers;
  int members;
};

#define CHANINV    0x0001
#define CHANPRIV   0x0002
#define CHANSEC    0x0004
#define CHANMODER  0x0008
#define CHANTOPIC  0x0010
#define CHANNOMSG  0x0020
#define CHANLIMIT  0x0040
#define CHANKEY    0x0080
#define CHANANON   0x0100
#define CHANQUIET  0x0200
#define CHANNOCLR  0x0400
#define CHANREGON  0x0800
#define CHANMODR   0x1000

struct chanset_t {
  struct chanset_t *next;
  struct chan_t channel;
  char dname[81];
  char name[81];
  char need_op[121];
  char need_key[121];
  char need_limit[121];
  char need_unban[121];
  char need_invite[121];
  int flood_pub_thr;
  int flood_pub_time;
  int flood_join_thr;
  int flood_join_time;
  int flood_deop_thr;
  int flood_deop_time;
  int flood_kick_thr;
  int flood_kick_time;
  int flood_ctcp_thr;
  int flood_ctcp_time;
  int flood_nick_thr;
  int flood_nick_time;
  int aop_min;
  int aop_max;
  int status;
  int ircnet_status;
  int idle_kick;
  int stopnethack_mode;
  int revenge_mode;
  maskrec *bans,
          *exempts,
          *invites;
  int mode_pls_prot;
  int mode_mns_prot;
  int limit_prot;
  char key_prot[121];
  char pls[21];
  char mns[21];
  char *key;
  char *rmkey;
  int limit;
  int bytes;
  int compat;
  struct {
    char *op;
    char type;
  } cmode[6];
  char floodwho[FLOOD_CHAN_MAX][81];
  time_t floodtime[FLOOD_CHAN_MAX];
  int floodnum[FLOOD_CHAN_MAX];
  char deopd[NICKLEN];
};

#define CHAN_ENFORCEBANS    0x0001
#define CHAN_DYNAMICBANS    0x0002
#define CHAN_NOUSERBANS     0x0004
#define CHAN_OPONJOIN       0x0008
#define CHAN_BITCH          0x0010
#define CHAN_GREET          0x0020
#define CHAN_PROTECTOPS     0x0040
#define CHAN_LOGSTATUS      0x0080
#define CHAN_REVENGE        0x0100
#define CHAN_SECRET         0x0200
#define CHAN_AUTOVOICE      0x0400
#define CHAN_CYCLE          0x0800
#define CHAN_DONTKICKOPS    0x1000
#define CHAN_INACTIVE       0x2000
#define CHAN_PROTECTFRIENDS 0x4000
#define CHAN_SHARED         0x8000
#define CHAN_SEEN           0x10000
#define CHAN_REVENGEBOT     0x20000
#define CHAN_NODESYNCH      0x40000
#define CHAN_ACTIVE         0x1000000
#define CHAN_PEND           0x2000000
#define CHAN_FLAGGED        0x4000000
#define CHAN_STATIC         0x8000000
#define CHAN_ASKEDBANS      0x10000000
#define CHAN_ASKEDMODES     0x20000000
#define CHAN_JUPED          0x40000000
#define CHAN_STOP_CYCLE     0x80000000
#define CHAN_ASKED_EXEMPTS  0x0001
#define CHAN_ASKED_INVITED  0x0002
#define CHAN_DYNAMICEXEMPTS 0x0004
#define CHAN_NOUSEREXEMPTS  0x0008
#define CHAN_DYNAMICINVITES 0x0010
#define CHAN_NOUSERINVITES  0x0020

memberlist *ismember(struct chanset_t *, char *);
struct chanset_t *findchan(const char *name);
struct chanset_t *findchan_by_dname(const char *name);

#define channel_hidden(chan) (chan->channel.mode & (CHANPRIV | CHANSEC))
#define channel_optopic(chan) (chan->channel.mode & CHANTOPIC)
#define channel_active(chan)  (chan->status & CHAN_ACTIVE)
#define channel_pending(chan)  (chan->status & CHAN_PEND)
#define channel_bitch(chan) (chan->status & CHAN_BITCH)
#define channel_nodesynch(chan) (chan->status & CHAN_NODESYNCH)
#define channel_autoop(chan) (chan->status & CHAN_OPONJOIN)
#define channel_autovoice(chan) (chan->status & CHAN_AUTOVOICE)
#define channel_greet(chan) (chan->status & CHAN_GREET)
#define channel_logstatus(chan) (chan->status & CHAN_LOGSTATUS)
#define channel_enforcebans(chan) (chan->status & CHAN_ENFORCEBANS)
#define channel_revenge(chan) (chan->status & CHAN_REVENGE)
#define channel_dynamicbans(chan) (chan->status & CHAN_DYNAMICBANS)
#define channel_nouserbans(chan) (chan->status & CHAN_NOUSERBANS)
#define channel_protectops(chan) (chan->status & CHAN_PROTECTOPS)
#define channel_protectfriends(chan) (chan->status & CHAN_PROTECTFRIENDS)
#define channel_dontkickops(chan) (chan->status & CHAN_DONTKICKOPS)
#define channel_secret(chan) (chan->status & CHAN_SECRET)
#define channel_shared(chan) (chan->status & CHAN_SHARED)
#define channel_static(chan) (chan->status & CHAN_STATIC)
#define channel_cycle(chan) (chan->status & CHAN_CYCLE)
#define channel_seen(chan) (chan->status & CHAN_SEEN)
#define channel_inactive(chan) (chan->status & CHAN_INACTIVE)
#define channel_revengebot(chan) (chan->status & CHAN_REVENGEBOT)
#define channel_dynamicexempts(chan) (chan->ircnet_status & CHAN_DYNAMICEXEMPTS)
#define channel_nouserexempts(chan) (chan->ircnet_status & CHAN_NOUSEREXEMPTS)
#define channel_dynamicinvites(chan) (chan->ircnet_status & CHAN_DYNAMICINVITES)
#define channel_nouserinvites(chan) (chan->ircnet_status & CHAN_NOUSERINVITES)
#define channel_juped(chan) (chan->status & CHAN_JUPED)
#define channel_stop_cycle(chan) (chan->status & CHAN_STOP_CYCLE)

struct msgq_head {
  struct msgq *head;
  struct msgq *last;
  int tot;
  int warned;
};

struct msgq {
  struct msgq *next;
  int len;
  char *msg;
};

#endif
