#define MODULE_NAME "server"
#define MAKING_SERVER
#include "src/mod/module.h"
#include "server.h"

static Function *global = NULL;
static int ctcp_mode;
static int serv;
static int strict_host;
static char newserver[121];
static int newserverport;
static char newserverpass[121];
static time_t trying_server;
static int server_lag;
static char altnick[NICKLEN];
static char raltnick[NICKLEN];
static int curserv;
static int flud_thr;
static int flud_time;
static int flud_ctcp_thr;
static int flud_ctcp_time;
static char initserver[121];
static char botuserhost[121];
static int keepnick;
static int nick_juped = 0;
static int check_stoned;
static int serverror_quit;
static int quiet_reject;
static int waiting_for_awake;
static time_t server_online;
static time_t server_cycle_wait;
static char botrealname[121];
static int min_servs;
static int server_timeout;
static int never_give_up;
static int strict_servernames;
static struct server_list *serverlist;
static int cycle_time;
static int default_port;
static char oldnick[NICKLEN];
static int trigger_on_ignore;
static int answer_ctcp;
static int lowercase_ctcp;
static int check_mode_r;
static int net_type;
static char connectserver[121];
static int resolvserv;
static int double_mode;
static int double_server;
static int double_help;
static int double_warned;
static int lastpingtime;
static char stackablecmds[511];
static char stackable2cmds[511];
static time_t last_time;
static int nick_len;
static int kick_method;
static p_tcl_bind_list H_wall, H_raw, H_notc, H_msgm, H_msg, H_flud, H_ctcr, H_ctcp;
static void empty_msgq(void);
static void next_server(int *, char *, unsigned int *, char *);
static void disconnect_server(int);
static char *get_altbotnick(void);
static void msgq_clear(struct msgq_head *qh);

#include "servmsg.c"
#define MAXPENALTY 10
#define msgrate 2

static int maxqmsg;
static struct msgq_head mq, hq, modeq;
static int burst;

#include "tclserv.c"

static void deq_msg()
{
  struct msgq *q;
  static time_t last_time = 0;
  int ok = 0;

  if (((now - last_time) >= msgrate) || (now < last_time)) {
    last_time = now;
    if (burst > 0)
      burst--;
    ok = 1;
  }
  if (serv < 0)
    return;
  if (modeq.head) {
    while (modeq.head && (burst < 4)) {
      tputs(serv, modeq.head->msg, modeq.head->len);
      modeq.tot--;
      q = modeq.head->next;
      nfree(modeq.head->msg);
      nfree(modeq.head);
      modeq.head = q;
      burst++;
    }
    if (!modeq.head)
      modeq.last = 0;
    return;
  }
  if (burst > 1)
    return;
  if (mq.head) {
    burst++;
    tputs(serv, mq.head->msg, mq.head->len);
    mq.tot--;
    q = mq.head->next;
    nfree(mq.head->msg);
    nfree(mq.head);
    mq.head = q;
    if (!mq.head)
      mq.last = 0;
    return;
  }
  if (!hq.head || burst || !ok)
    return;
  tputs(serv, hq.head->msg, hq.head->len);
  hq.tot--;
  q = hq.head->next;
  nfree(hq.head->msg);
  nfree(hq.head);
  hq.head = q;
  if (!hq.head)
    hq.last = 0;
}

static void empty_msgq()
{
  msgq_clear(&modeq);
  msgq_clear(&mq);
  msgq_clear(&hq);
  burst = 0;
}

static void queue_server(int which, char *buf, int len)
{
  struct msgq_head *h = 0;
  struct msgq *q;
  struct msgq_head tempq;
  struct msgq *tq, *tqq;
  int doublemsg;
  doublemsg = 0;
  if (serv < 0)
    return;			
      if ((strncasecmp(buf, "PING", 4) == 0) || (strncasecmp(buf, "PONG", 4) == 0)) {
    if ((buf[1] == 73) || (buf[1] == 105)) {
      lastpingtime = now;
    }
    tputs(serv, buf, len);
    return;
  }
  switch (which) {
  case DP_MODE:
    h = &modeq;
    tempq = modeq;
    if (double_mode) doublemsg = 1;
    break;
  case DP_SERVER:
    h = &mq;
    tempq = mq;
    if (double_server) doublemsg = 1;
    break;
  case DP_HELP:
    h = &hq;
    tempq = hq;
    if (double_help) doublemsg = 1;
    break;
  default:
    return;
  }
  if (h->tot < maxqmsg) {
    if (!doublemsg) {
      tq = tempq.head;
      while (tq) {
	tqq = tq->next;
	if (!strcasecmp(tq->msg, buf)) {
	  if (!double_warned) {
	    if (buf[len - 1] == '\n')
	      buf[len - 1] = 0;
	    debug1("msg already queued. skipping: %s", buf);
	    double_warned = 1;
	  }
	  return;
	}
	tq = tqq;
      }
    }
    q = nmalloc(sizeof(struct msgq));
    q->next = NULL;
    if (h->head)
      h->last->next = q;
    else
      h->head = q;
    h->last = q;
    q->len = len;
    q->msg = nmalloc(len + 1);
    strcpy(q->msg, buf);
    h->tot++;
    h->warned = 0;
    double_warned = 0;
  } else {
    h->warned = 1;
  }
  if (debug_output) {
    if (buf[len - 1] == '\n')
      buf[len - 1] = 0;
    switch (which) {
    case DP_MODE:
      break;
    case DP_SERVER:
      break;
    case DP_HELP:
      break;
    }
  }
  if (which == DP_MODE)
    deq_msg();
}

static void add_server(char *ss)
{
  struct server_list *x, *z;
  char *p, *q;

  for (z = serverlist; z && z->next; z = z->next);
  while (ss) {
    p = strchr(ss, ',');
    if (p)
      *p++ = 0;
    x = nmalloc(sizeof(struct server_list));

    x->next = 0;
    x->realname = 0;
    if (z)
      z->next = x;
    else
      serverlist = x;
    z = x;
    q = strchr(ss, ':');
    if (!q) {
      x->port = default_port;
      x->pass = 0;
      x->name = nmalloc(strlen(ss) + 1);
      strcpy(x->name, ss);
    } else {
      *q++ = 0;
      x->name = nmalloc(q - ss);
      strcpy(x->name, ss);
      ss = q;
      q = strchr(ss, ':');
      if (!q) {
	x->pass = 0;
      } else {
	*q++ = 0;
	x->pass = nmalloc(strlen(q) + 1);
	strcpy(x->pass, q);
      }
      x->port = atoi(ss);
    }
    ss = p;
  }
}

static void clearq(struct server_list *xx)
{
  struct server_list *x;

  while (xx) {
    x = xx->next;
    if (xx->name)
      nfree(xx->name);
    if (xx->pass)
      nfree(xx->pass);
    if (xx->realname)
      nfree(xx->realname);
    nfree(xx);
    xx = x;
  }
}

static void next_server(int *ptr, char *serv, unsigned int *port, char *pass)
{
  struct server_list *x = serverlist;
  int i = 0;

  if (x == NULL)
    return;
  if (*ptr == (-1)) {
    for (; x; x = x->next) {
      if (x->port == *port) {
	if (!egg_strcasecmp(x->name, serv)) {
	  *ptr = i;
	  return;
	} else if (x->realname && !egg_strcasecmp(x->realname, serv)) {
	  *ptr = i;
	  strncpyz(serv, x->realname, sizeof serv);
	  return;
	}
      }
      i++;
    }
    x = nmalloc(sizeof(struct server_list));
    x->next = 0;
    x->realname = 0;
    x->name = nmalloc(strlen(serv) + 1);
    strcpy(x->name, serv);
    x->port = *port ? *port : default_port;
    if (pass && pass[0]) {
      x->pass = nmalloc(strlen(pass) + 1);
      strcpy(x->pass, pass);
    } else
      x->pass = NULL;
    list_append((struct list_type **) (&serverlist), (struct list_type *) x);
    *ptr = i;
    return;
  }
  i = (*ptr);
  while (i > 0 && x != NULL) {
    x = x->next;
    i--;
  }
  if (x != NULL) {
    x = x->next;
    (*ptr)++;
  }
  if (x == NULL) {
    x = serverlist;
    *ptr = 0;
  }
  strcpy(serv, x->name);
  *port = x->port ? x->port : default_port;
  if (x->pass)
    strcpy(pass, x->pass);
  else
    pass[0] = 0;
}

static int server_6char STDVAR
{
  Function F = (Function) cd;
  char x[20];

  BADARGS(7, 7, " nick user@host handle desto/chan keyword/nick text");
  CHECKVALIDITY(server_6char);
  egg_snprintf(x, sizeof x, "%d",
	       F(argv[1], argv[2], argv[3], argv[4], argv[5], argv[6]));
  Tcl_AppendResult(irp, x, NULL);
  return TCL_OK;
}

static int server_5char STDVAR
{
  Function F = (Function) cd;

  BADARGS(6, 6, " nick user@host handle channel text");
  CHECKVALIDITY(server_5char);
  F(argv[1], argv[2], argv[3], argv[4], argv[5]);
  return TCL_OK;
}

static int server_2char STDVAR
{
  Function F = (Function) cd;

  BADARGS(3, 3, " nick msg");
  CHECKVALIDITY(server_2char);
  F(argv[1], argv[2]);
  return TCL_OK;
}

static int server_msg STDVAR
{
  Function F = (Function) cd;

  BADARGS(5, 5, " nick uhost hand buffer");
  CHECKVALIDITY(server_msg);
  F(argv[1], argv[2], get_user_by_handle(userlist, argv[3]), argv[4]);
  return TCL_OK;
}

static int server_raw STDVAR
{
  Function F = (Function) cd;

  BADARGS(4, 4, " from code args");
  CHECKVALIDITY(server_raw);
  Tcl_AppendResult(irp, int_to_base10(F(argv[1], argv[3])), NULL);
  return TCL_OK;
}

static char *nick_change(ClientData cdata, Tcl_Interp *irp, char *name1,
			 char *name2, int flags)
{
  char *new;

  if (flags & (TCL_TRACE_READS | TCL_TRACE_UNSETS)) {
    Tcl_SetVar2(interp, name1, name2, origbotname, TCL_GLOBAL_ONLY);
    if (flags & TCL_TRACE_UNSETS)
      Tcl_TraceVar(irp, name1, TCL_TRACE_READS | TCL_TRACE_WRITES |
        	   TCL_TRACE_UNSETS, nick_change, cdata);
  } else {
    new = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
    if (rfc_casecmp(origbotname, new)) {
      if (origbotname[0]) {
        nick_juped = 0;
      }
      strncpyz(origbotname, new, NICKLEN);
      if (server_online)
	dprintf(DP_SERVER, "NICK %s\n", origbotname);
    }
  }
  return NULL;
}

static void rand_nick(char *nick)
{
  register char *p = nick;

  while ((p = strchr(p, '?')) != NULL) {
    *p = '0' + random() % 10;
    p++;
  }
}

static char *get_altbotnick(void)
{
  if (strchr(altnick, '?')) {
    if (!raltnick[0]) {
      strncpyz(raltnick, altnick, NICKLEN);
      rand_nick(raltnick);
    }
    return raltnick;
  } else
    return altnick;
}

static char *altnick_change(ClientData cdata, Tcl_Interp *irp, char *name1,
			    char *name2, int flags)
{
  raltnick[0] = 0;
  return NULL;
}

static char *traced_server(ClientData cdata, Tcl_Interp *irp, char *name1,
			   char *name2, int flags)
{
  char s[1024];

  if (server_online) {
    int servidx = findanyidx(serv);

    simple_sprintf(s, "%s:%u", dcc[servidx].host, dcc[servidx].port);
  } else
    s[0] = 0;
  Tcl_SetVar2(interp, name1, name2, s, TCL_GLOBAL_ONLY);
  if (flags & TCL_TRACE_UNSETS)
    Tcl_TraceVar(irp, name1, TCL_TRACE_READS | TCL_TRACE_WRITES |
		 TCL_TRACE_UNSETS, traced_server, cdata);
  return NULL;
}

static char *traced_botname(ClientData cdata, Tcl_Interp *irp, char *name1,
			    char *name2, int flags)
{
  char s[1024];

  simple_sprintf(s, "%s%s%s", botname, 
		 botuserhost[0] ? "!" : "", botuserhost[0] ? botuserhost : "");
  Tcl_SetVar2(interp, name1, name2, s, TCL_GLOBAL_ONLY);
  if (flags & TCL_TRACE_UNSETS)
    Tcl_TraceVar(irp, name1, TCL_TRACE_READS | TCL_TRACE_WRITES |
		 TCL_TRACE_UNSETS, traced_botname, cdata);
  return NULL;
}

static void do_nettype(void)
{
  switch (net_type) {
  case NETT_EFNET:
    check_mode_r = 0;
    nick_len = 9;
    break;
  case NETT_IRCNET:
    check_mode_r = 1;
    nick_len = 9;
    simple_sprintf(stackablecmds, "INVITE AWAY VERSION NICK");
    kick_method = 4;
    break;
  case NETT_UNDERNET:
    check_mode_r = 0;
    nick_len = 9;
    simple_sprintf(stackablecmds, "PRIVMSG NOTICE TOPIC PART WHOIS USERHOST USERIP ISON");
    simple_sprintf(stackable2cmds, "USERHOST USERIP ISON");
    break;
  case NETT_DALNET:
    check_mode_r = 0;
    nick_len = 32;
    simple_sprintf(stackablecmds, "PRIVMSG NOTICE PART WHOIS WHOWAS USERHOST ISON WATCH DCCALLOW");
    simple_sprintf(stackable2cmds, "USERHOST ISON WATCH");
    break;
  case NETT_HYBRID_EFNET:
    check_mode_r = 0;
    nick_len = 9;
    break;
  }
}

static char *traced_nettype(ClientData cdata, Tcl_Interp *irp, char *name1,
			    char *name2, int flags)
{
  do_nettype();
  return NULL;
}

static char *traced_nicklen(ClientData cdata, Tcl_Interp *irp, char *name1,
			    char *name2, int flags)
{
  if (flags & (TCL_TRACE_READS | TCL_TRACE_UNSETS)) {
    char s[40];

    egg_snprintf(s, sizeof s, "%d", nick_len);
    Tcl_SetVar2(interp, name1, name2, s, TCL_GLOBAL_ONLY);
    if (flags & TCL_TRACE_UNSETS)
      Tcl_TraceVar(irp, name1, TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		   traced_nicklen, cdata);
  } else {
    char *cval = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
    long lval = 0;

    if (cval && Tcl_ExprLong(interp, cval, &lval) != TCL_ERROR) {
      if (lval > NICKMAX)
	lval = NICKMAX;
      nick_len = (int) lval;
    }
  }
  return NULL;
}

static tcl_strings my_tcl_strings[] =
{
  {"botnick",			NULL,		0,		STR_PROTECT},
  {"altnick",			altnick,	NICKMAX,	0},
  {"realname",			botrealname,	80,		0},
  {"init-server",		initserver,	120,		0},
  {"connect-server",		connectserver,	120,		0},
  {"stackable-commands",	stackablecmds,	510,		0},
  {"stackable2-commands",	stackable2cmds,	510,		0},
  {NULL,			NULL,		0,		0}
};

static tcl_coups my_tcl_coups[] =
{
  {"flood-msg",		&flud_thr,		&flud_time},
  {"flood-ctcp",	&flud_ctcp_thr, 	&flud_ctcp_time},
  {NULL,		NULL,			NULL}
};

static tcl_ints my_tcl_ints[] =
{
  {"use-console-r",		NULL,				1},
  {"servlimit",			&min_servs,			0},
  {"server-timeout",		&server_timeout,		0},
  {"lowercase-ctcp",		&lowercase_ctcp,		0},
  {"server-online",		(int *) &server_online,		2},
  {"never-give-up",		&never_give_up,			0},
  {"keep-nick",			&keepnick,			0},
  {"strict-servernames",	&strict_servernames,		0},
  {"check-stoned",		&check_stoned,			0},
  {"serverror-quit",		&serverror_quit,		0},
  {"quiet-reject",		&quiet_reject,			1},
  {"max-queue-msg",		&maxqmsg,			0},
  {"trigger-on-ignore",		&trigger_on_ignore,		0},
  {"answer-ctcp",		&answer_ctcp,			0},
  {"server-cycle-wait",		(int *) &server_cycle_wait,	0},
  {"default-port",		&default_port,			0},
  {"check-mode-r",		&check_mode_r,			0},
  {"net-type",			&net_type,			0},
  {"ctcp-mode",			&ctcp_mode,			0},
  {"double-mode",		&double_mode,			0},
  {"double-server",		&double_server,			0},
  {"double-help",		&double_help,			0},
  {"nicklen",			&nick_len,			0},
  {"nick-len",			&nick_len,			0},
  {"isjuped",			&nick_juped,			0},
  {NULL,			NULL,				0}
};

static char *tcl_eggserver(ClientData cdata, Tcl_Interp *irp, char *name1,
			   char *name2, int flags)
{
  Tcl_DString ds;
  char *slist, **list, x[1024];
  struct server_list *q;
  int lc, code, i;

  if (flags & (TCL_TRACE_READS | TCL_TRACE_UNSETS)) {
    Tcl_DStringInit(&ds);
    for (q = serverlist; q; q = q->next) {
      egg_snprintf(x, sizeof x, "%s:%d%s%s %s", q->name,
		   q->port ? q->port : default_port, q->pass ? ":" : "",
		   q->pass ? q->pass : "", q->realname ? q->realname : "");
      Tcl_DStringAppendElement(&ds, x);
    }
    slist = Tcl_DStringValue(&ds);
    Tcl_SetVar2(interp, name1, name2, slist, TCL_GLOBAL_ONLY);
    Tcl_DStringFree(&ds);
  } else {
    if (serverlist) {
      clearq(serverlist);
      serverlist = NULL;
    }
    slist = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
    if (slist != NULL) {
      code = Tcl_SplitList(interp, slist, &lc, &list);
      if (code == TCL_ERROR)
	return interp->result;
      for (i = 0; i < lc && i < 50; i++)
	add_server(list[i]);
      if (server_online) {
	int servidx = findanyidx(serv);

	curserv = (-1);
	next_server(&curserv, dcc[servidx].host, &dcc[servidx].port, "");
      }
      Tcl_Free((char *) list);
    }
  }
  return NULL;
}

#define tcl_traceserver(name, ptr) \
  Tcl_TraceVar(interp, name, TCL_TRACE_READS | TCL_TRACE_WRITES |	\
	       TCL_TRACE_UNSETS, tcl_eggserver, (ClientData) ptr)
#define tcl_untraceserver(name, ptr) \
  Tcl_UntraceVar(interp, name, TCL_TRACE_READS | TCL_TRACE_WRITES |	\
		 TCL_TRACE_UNSETS, tcl_eggserver, (ClientData) ptr)

static void server_secondly()
{
  if (cycle_time)
    cycle_time--;
  deq_msg();
  if (!resolvserv && serv < 0)
    connect_server();
}

static void server_5minutely()
{
  if (check_stoned) {
    if (waiting_for_awake) {
      int servidx = findanyidx(serv);

      disconnect_server(servidx);
      lostdcc(servidx);
    } else if (!trying_server) {
      dprintf(DP_MODE, "PING :%lu\n", (unsigned long) now);
      waiting_for_awake = 1;
    }
  }
}

static void server_prerehash()
{
  strcpy(oldnick, botname);
}

static void server_postrehash()
{
  strncpyz(botname, origbotname, NICKLEN);
  if (!botname[0])
    fatal("NO BOT NAME.", 0);
  if (serverlist == NULL)
    fatal("NO SERVER.", 0);
    if (oldnick[0] && !rfc_casecmp(oldnick, botname)
       && !rfc_casecmp(oldnick, get_altbotnick())) {
    strcpy(botname, oldnick);
    dprintf(DP_SERVER, "NICK %s\n", origbotname);
  }
  else if (oldnick[0])
    strcpy(botname, oldnick);
  if (initserver[0])
    do_tcl("init-server", initserver);
}

static void server_die()
{
  cycle_time = 100;
  if (server_online) {
    dprintf(-serv, "QUIT :%s\n", quit_msg[0] ? quit_msg : "");
    sleep(3);
  }
  nuke_server(NULL);
}

static void server_report(int idx, int details)
{
  char s1[64], s[128];
  int servidx;

  if (server_online) {
    dprintf(idx, "    Online as: %s%s%s (%s)\n", botname,
	    botuserhost[0] ? "!" : "", botuserhost[0] ? botuserhost : "",
	    botrealname);
    if (nick_juped)
      dprintf(idx, "    NICK IS JUPED: %s %s\n", origbotname,
	      keepnick ? "(trying)" : "");
    nick_juped = 0;
    daysdur(now, server_online, s1);
    egg_snprintf(s, sizeof s, "(connected %s)", s1);
    if (server_lag && !waiting_for_awake) {
      if (server_lag == (-1))
	egg_snprintf(s1, sizeof s1, " (bad pong replies)");
      else
	egg_snprintf(s1, sizeof s1, " (lag: %ds)", server_lag);
      strcat(s, s1);
    }
  }
  if ((trying_server || server_online) &&
	((servidx = findanyidx(serv)) != (-1))) {
    dprintf(idx, "    Server %s:%d %s\n", dcc[servidx].host, dcc[servidx].port,
	    trying_server ? "(trying)" : s);
  } else
    dprintf(idx, "    %s\n", IRC_NOSERVER);
  if (modeq.tot)
    dprintf(idx, "    %s %d%%, %d msgs\n", IRC_MODEQUEUE,
            (int) ((float) (modeq.tot * 100.0) / (float) maxqmsg),
	    (int) modeq.tot);
  if (mq.tot)
    dprintf(idx, "    %s %d%%, %d msgs\n", IRC_SERVERQUEUE,
           (int) ((float) (mq.tot * 100.0) / (float) maxqmsg), (int) mq.tot);
  if (hq.tot)
    dprintf(idx, "    %s %d%%, %d msgs\n", IRC_HELPQUEUE,
           (int) ((float) (hq.tot * 100.0) / (float) maxqmsg), (int) hq.tot);
  if (details) {
    if (min_servs)
      dprintf(idx, "    Requiring a net of at least %d server(s)\n", min_servs);
    if (initserver[0])
      dprintf(idx, "    On connect, I do: %s\n", initserver);
    if (connectserver[0])
      dprintf(idx, "    Before connect, I do: %s\n", connectserver);
    dprintf(idx, "    Flood is: %d msg/%ds, %d ctcp/%ds\n",
	    flud_thr, flud_time, flud_ctcp_thr, flud_ctcp_time);
  }
}

static void msgq_clear(struct msgq_head *qh)
{
  register struct msgq	*q, *qq;

  for (q = qh->head; q; q = qq) {
    qq = q->next;
    nfree(q->msg);
    nfree(q);
  }
  qh->head = qh->last = NULL;
  qh->tot = qh->warned = 0;
}

static int msgq_expmem(struct msgq_head *qh)
{
  register int		 tot = 0;
  register struct msgq	*m;

  for (m = qh->head; m; m = m->next) {
    tot += m->len + 1;
    tot += sizeof(struct msgq);
  }
  return tot;
}

static int server_expmem()
{
  int			 tot = 0;
  struct server_list	*s = serverlist;

  for (; s; s = s->next) {
    if (s->name)
      tot += strlen(s->name) + 1;
    if (s->pass)
      tot += strlen(s->pass) + 1;
    if (s->realname)
      tot += strlen(s->realname) + 1;
    tot += sizeof(struct server_list);
  }

  tot += msgq_expmem(&mq) + msgq_expmem(&hq) + msgq_expmem(&modeq);

  return tot;
}

static cmd_t my_ctcps[] =
{
  {NULL,	NULL,	NULL,			NULL}
};

static char *server_close()
{
  cycle_time = 100;
  nuke_server("Connection reset by peer");
  clearq(serverlist);
  rem_builtins(H_raw, my_raw_binds);
  rem_builtins(H_ctcp, my_ctcps);
  del_bind_table(H_wall);
  del_bind_table(H_raw);
  del_bind_table(H_notc);
  del_bind_table(H_msgm);
  del_bind_table(H_msg);
  del_bind_table(H_flud);
  del_bind_table(H_ctcr);
  del_bind_table(H_ctcp);
  rem_tcl_coups(my_tcl_coups);
  rem_tcl_strings(my_tcl_strings);
  rem_tcl_ints(my_tcl_ints);
  rem_tcl_commands(my_tcl_cmds);
  Tcl_UntraceVar(interp, "nick",
		 TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		 nick_change, NULL);
  Tcl_UntraceVar(interp, "altnick",
		 TCL_TRACE_WRITES | TCL_TRACE_UNSETS, altnick_change, NULL);
  Tcl_UntraceVar(interp, "botname",
		 TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		 traced_botname, NULL);
  Tcl_UntraceVar(interp, "server",
		 TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		 traced_server, NULL);
  Tcl_UntraceVar(interp, "net-type",
		 TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		 traced_nettype, NULL);
  Tcl_UntraceVar(interp, "nick-len",
		 TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		 traced_nicklen, NULL);
  tcl_untraceserver("servers", NULL);
  empty_msgq();
  del_hook(HOOK_SECONDLY, (Function) server_secondly);
  del_hook(HOOK_5MINUTELY, (Function) server_5minutely);
  del_hook(HOOK_QSERV, (Function) queue_server);
  del_hook(HOOK_MINUTELY, (Function) minutely_checks);
  del_hook(HOOK_PRE_REHASH, (Function) server_prerehash);
  del_hook(HOOK_REHASH, (Function) server_postrehash);
  del_hook(HOOK_DIE, (Function) server_die);
  module_undepend(MODULE_NAME);
  return NULL;
}

EXPORT_SCOPE char *server_start();

static Function server_table[] =
{
  (Function) server_start,
  (Function) server_close,
  (Function) server_expmem,
  (Function) server_report,
  (Function) NULL,
  (Function) botuserhost,
  (Function) & quiet_reject,
  (Function) & serv,
  (Function) & flud_thr,
  (Function) & flud_time,
  (Function) & flud_ctcp_thr,
  (Function) & flud_ctcp_time,
  (Function) match_my_nick,
  (Function) check_tcl_flud,
  (Function) NULL,
  (Function) & answer_ctcp,
  (Function) & trigger_on_ignore,
  (Function) check_tcl_ctcpr,
  (Function) detect_avalanche,
  (Function) nuke_server,
  (Function) newserver,	
  (Function) & newserverport,
  (Function) newserverpass,
  (Function) & cycle_time,
  (Function) & default_port,
  (Function) & server_online,
  (Function) & min_servs,
  (Function) & H_raw,
  (Function) & H_wall,
  (Function) & H_msg,
  (Function) & H_msgm,
  (Function) & H_notc,
  (Function) & H_flud,
  (Function) & H_ctcp,
  (Function) & H_ctcr,
  (Function) ctcp_reply,
  (Function) get_altbotnick,
  (Function) & nick_len,
  (Function) check_tcl_notc
};

char *server_start(Function *global_funcs)
{
  char *s;
  global = global_funcs;

  serv = -1;
  strict_host = 1;
  botname[0] = 0;
  trying_server = 0L;
  server_lag = 0;
  altnick[0] = 0;
  raltnick[0] = 0;
  curserv = 0;
  flud_thr = 0;
  flud_time = 0;
  flud_ctcp_thr = 0;
  flud_ctcp_time = 0;
  initserver[0] = 0;
  connectserver[0] = 0;
  botuserhost[0] = 0;
  keepnick = 1;
  check_stoned = 1;
  serverror_quit = 1;
  quiet_reject = 1;
  waiting_for_awake = 0;
  server_online = 0;
  server_cycle_wait = 60;
  strcpy(botrealname, "A deranged product of suid coders");
  min_servs = 0;
  server_timeout = 320;
  never_give_up = 0;
  strict_servernames = 0;
  serverlist = NULL;
  cycle_time = 0;
  default_port = 6667;
  oldnick[0] = 0;
  trigger_on_ignore = 0;
  answer_ctcp = 0;
  lowercase_ctcp = 0;
  check_mode_r = 0;
  maxqmsg = 300;
  burst = 0;
  net_type = 1;
  double_mode = 0;
  double_server = 0;
  double_help = 0;
  stackablecmds[0] = 0;
  strcpy(stackable2cmds, "USERHOST ISON");
  resolvserv = 0;
  lastpingtime = 0;
  last_time = 0;
  nick_len = 9;
  kick_method = 4;

  server_table[4] = (Function) botname;
  module_register(MODULE_NAME, server_table, 1, 2);
  if (!module_depend(MODULE_NAME, "suid", 304, 0)) {
    module_undepend(MODULE_NAME);
    return "This module requires suid3.6 or later";
  }

  tcl_eggserver(NULL, interp, "servers", NULL, 0);
  tcl_traceserver("servers", NULL);
  s = Tcl_GetVar(interp, "nick", TCL_GLOBAL_ONLY);
  if (s)
    strncpyz(origbotname, s, NICKLEN);
  Tcl_TraceVar(interp, "nick",
	       TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
	       nick_change, NULL);
  Tcl_TraceVar(interp, "altnick",
	       TCL_TRACE_WRITES | TCL_TRACE_UNSETS, altnick_change, NULL);
  Tcl_TraceVar(interp, "botname",
	       TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
	       traced_botname, NULL);
  Tcl_TraceVar(interp, "server",
	       TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
	       traced_server, NULL);
  Tcl_TraceVar(interp, "net-type",
	       TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
	       traced_nettype, NULL);
  Tcl_TraceVar(interp, "nick-len",
	       TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
	       traced_nicklen, NULL);

  H_wall = add_bind_table("wall", HT_STACKABLE, server_2char);
  H_raw = add_bind_table("raw", HT_STACKABLE, server_raw);
  H_notc = add_bind_table("notc", HT_STACKABLE, server_6char);
  H_msgm = add_bind_table("msgm", HT_STACKABLE, server_msg);
  H_msg = add_bind_table("msg", 0, server_msg);
  H_flud = add_bind_table("flud", HT_STACKABLE, server_5char);
  H_ctcr = add_bind_table("ctcr", HT_STACKABLE, server_6char);
  H_ctcp = add_bind_table("ctcp", HT_STACKABLE, server_6char);
  add_builtins(H_raw, my_raw_binds);
  add_builtins(H_ctcp, my_ctcps);
  my_tcl_strings[0].buf = botname;
  add_tcl_strings(my_tcl_strings);
  my_tcl_ints[0].val = &use_console_r;
  add_tcl_ints(my_tcl_ints);
  add_tcl_commands(my_tcl_cmds);
  add_tcl_coups(my_tcl_coups);
  add_hook(HOOK_SECONDLY, (Function) server_secondly);
  add_hook(HOOK_5MINUTELY, (Function) server_5minutely);
  add_hook(HOOK_MINUTELY, (Function) minutely_checks);
  add_hook(HOOK_QSERV, (Function) queue_server);
  add_hook(HOOK_PRE_REHASH, (Function) server_prerehash);
  add_hook(HOOK_REHASH, (Function) server_postrehash);
  add_hook(HOOK_DIE, (Function) server_die);
  mq.head = hq.head = modeq.head = NULL;
  mq.last = hq.last = modeq.last = NULL;
  mq.tot = hq.tot = modeq.tot = 0;
  mq.warned = hq.warned = modeq.warned = 0;
  double_warned = 0;
  newserver[0] = 0;
  newserverport = 0;
  curserv = 999;
  do_nettype();
  return NULL;
}
